
(****************************************************************
   File: miniml.v                                                 
   Authors: Amy Felty
   Version: Coq V8.0pl3
   Date: May 2007
                                                                 
  ***************************************************************)

Require Import Hybrid.
Require Export Omega.

Section miniml.

Hint Rewrite lbind_CON lbind_id lbind_VAR lbind_BND
             lbind_APP lbind_ABS : lbind_rw.

(****************************************************************
   Representing Lambda Terms as Constants
  ****************************************************************)

Inductive Lcon: Set :=
  cABS: Lcon | cAPP: Lcon | cFIX : Lcon.
Definition uexp : Set := expr Lcon.

Definition Var : var -> uexp := (VAR Lcon).
Definition Bnd : bnd -> uexp := (BND Lcon).
Definition App : uexp -> uexp -> uexp :=
  fun e1:uexp => fun e2:uexp => (APP (APP (CON cAPP) e1) e2).
Definition Fun : (uexp -> uexp) -> uexp :=
  fun f:uexp->uexp => (APP (CON cABS) (lambda f)).
Definition Fix : (uexp -> uexp) -> uexp :=
  fun f:uexp->uexp => (APP (CON cFIX) (lambda f)).

Inductive isterm1 : list uexp -> uexp -> Prop :=
| term1_var : forall (G:list uexp) (x:uexp), In x G -> isterm1 G x
| term1_app : forall (G:list uexp) (M N:uexp),
              isterm1 G M -> isterm1 G N -> isterm1 G (App M N)
| term1_lam : forall (G:list uexp) (E:uexp -> uexp), abstr E ->
              (forall x:uexp, proper x -> isterm1 (x::G) (E x)) ->
              isterm1 G (Fun (fun x => (E x)))
| term1_fix : forall (G:list uexp) (E:uexp -> uexp), abstr E ->
              (forall x:uexp, proper x -> isterm1 (x::G) (E x)) ->
              isterm1 G (Fix (fun x => (E x))).

(****************************************************************
   Types
  ****************************************************************)

Inductive tp: Set :=
   gnd : tp
 | arrow : tp -> tp -> tp.

(****************************************************************
   Example
  ****************************************************************)

Definition example1 : uexp :=
  (Fix (fun x => (Fun (fun y => (App x y))))).

Definition example2 : uexp :=
  (Fun (fun x => (Fun (fun y => x)))).

Definition example3 : uexp :=
  (Fun (fun x => (Fun (fun y => (App x y))))).

(* should be provable *)
Lemma isterm1_ex1 : isterm1 nil example1.
Proof.
unfold example1.
generalize (term1_fix nil (fun x : uexp => Fun (fun y : uexp => App x y)));
  intro h1.
apply h1; clear h1; auto.
Focus 2.
intros x h1.
generalize (term1_lam (x::nil) (fun y : uexp => App x y)); intro h2.
apply h2; auto.
unfold abstr.
unfold abst.
exists (fun y:uexp => App x y); split; auto with Ext_Eq.
generalize abst_APP; intro H.
generalize (H Lcon 0); clear H; intro H.
generalize (H (fun y:uexp => x) (fun y:uexp => y)); clear h2 H; intro h2.
unfold App.
Abort.

Lemma eq1_ex1 :
 example1 = (APP (CON cFIX)
             (lambda
              (fun x : uexp =>
               APP (CON cABS)
                (lambda (fun y : uexp => APP (APP (CON cAPP) x) y))))).
Proof.
unfold example1.
unfold Fix,Fun,App; auto.
Qed.

Lemma eq2_ex1 :
 example1 = (APP (CON cFIX)
             (lambda
              (fun x : uexp =>
               APP (CON cABS)
                (lambda (fun y : uexp => APP (APP (CON cAPP) x) y))))).
Proof.
unfold example1.
unfold Fix,Fun,App.
unfold lambda.
generalize (lbind_APP 0 (fun x => (CON cABS))
            (fun x => (ABS (lbind 0 (fun y => APP (APP (CON cAPP) x) y)))));
  intro H.
unfold uexp; rewrite H; clear H.
autorewrite with lbind_rw.
generalize (lbind_ABS 0 (fun x => (lbind 0 (fun y =>
             APP (APP (CON cAPP) x) y)))); intro H.
rewrite H; clear H.
auto.
Qed.

Lemma test2_abstr :
  abstr (fun x : uexp => Fun (fun _ : uexp => x)).
Proof.
unfold abstr, abst.
unfold Fun, lambda.
exists (fun x => APP (CON cABS) (ABS x)); split.
unfold ext_eq.
intro x.
specialize (lbind_does_not_occur x 0); intro H.
unfold uexp; rewrite H; auto.
generalize abst_APP; intro H.
generalize (H Lcon 0); clear H; intro H.
generalize (H (fun x:uexp => (CON cABS))); clear H; intro H.
generalize (H (fun x:uexp => (ABS x))); clear H; intro H.
apply H; clear H; auto.
unfold uexp; apply abst_CON; auto.
generalize abst_ABS; intro H.
generalize (H Lcon 0); clear H; intro H.
generalize (H (fun x => x)); clear H; intro H.
apply H; clear H; auto.
apply abst_id.
Qed.

Lemma test3a_abstr : forall x:uexp, proper x ->
  abstr (fun y : uexp => App x y).
Proof.
intros x H.
unfold abstr, abst.
exists (fun y : uexp => App x y); split.
unfold ext_eq; auto.
unfold App.
generalize abst_APP; intro H'.
generalize (H' Lcon 0); clear H'; intro H'.
generalize (H' (fun y:uexp => (APP (CON cAPP) x))); clear H'; intro H'.
generalize (H' (fun y:uexp => y)); clear H'; intro H'.
apply H'.
generalize abst_APP; intro H''.
generalize (H'' Lcon 0); clear H''; intro H''.
generalize (H'' (fun _ => (CON cAPP))); clear H''; intro H''.
generalize (H'' (fun _ => x)); clear H''; intro H''.
apply H''; clear H''; auto.
apply abst_CON; auto.
generalize proper_abst_aux.
intro H0.
generalize (H0 Lcon x H); auto.
unfold uexp; apply abst_id.
Qed.

(****************************************************************
   Some Properties of Constructors
  ****************************************************************)

Lemma proper_Var: forall x:var, (proper (Var x)).
Proof.
unfold proper,Var; apply level_VAR.
Qed.

Lemma proper_App: forall (E1 E2:uexp),
  proper E1 -> proper E2 -> proper (App E1 E2).
Proof.
unfold App,proper.
intros E1 E2 h1 h2.
apply level_APP; auto.
apply level_APP; auto.
apply level_CON.
Qed.

Lemma proper_App_l : forall (E1 E2:uexp),
  proper (App E1 E2) -> proper E1.
Proof.
intros E1 E2 h.
inversion h; subst; clear h.
inversion H2; subst; auto.
Qed.

Lemma proper_App_r : forall (E1 E2:uexp),
  proper (App E1 E2) -> proper E2.
Proof.
intros E1 E2 h.
inversion h; subst; clear h; auto.
Qed.

Lemma proper_Fun: forall (E:uexp -> uexp),
  abstr E -> proper (Fun (fun x => E x)).
Proof.
intros E h.
unfold Fun,proper.
apply level_APP.
apply level_CON.
apply abstr_proper.
specialize abstr_eta with (1:=h); intro h1.
auto.
Qed.

Lemma proper_Fix:
  forall (E:uexp -> uexp), abstr E -> proper (Fix (fun x => E x)).
Proof.
intros E h.
unfold Fix,proper.
apply level_APP.
apply level_CON.
apply abstr_proper.
specialize abstr_eta with (1:=h); intro h1.
auto.
Qed.

Lemma abstr_proper_Fix:
  forall E:uexp->uexp, (abstr E) -> (proper (Fix E)).
Proof.
intros E h.
unfold Fix,proper.
apply level_APP.
apply level_CON.
apply abstr_proper.
auto.
Qed.

Lemma FA_clash : forall (e:uexp->uexp) (e1 e2:uexp), (Fun e)<>(App e1 e1).
Proof.
intros e e1 e2 H.
inversion H.
Qed.

Lemma Fun_inj: forall E E':uexp->uexp,
  (abstr E) -> (abstr E') -> (Fun E)=(Fun E')->(ext_eq E E').
Proof.
unfold Fun; intros E E' H H0 H1.
inversion H1.
apply abstr_lbind_simp with 0; auto.
Qed.

Lemma Fix_inj: forall E E':uexp->uexp,
  (abstr E) -> (abstr E') -> (Fix E)=(Fix E')->(ext_eq E E').
Proof.
unfold Fix; intros E E' H H0 H1.
inversion H1.
apply abstr_lbind_simp with 0; auto.
Qed.

Lemma clash_App_Fun : forall (e:uexp->uexp) (e1 e2:uexp) (P:Prop),
  (Fun e)=(App e1 e1) -> P.
Proof.
intros e e1 e2 P H.
generalize (FA_clash e e1 e2); intro H0; elim H0; auto.
Qed.

End miniml.
