
(****************************************************************
   File: spec.v                                                 
   Authors: Amy Felty
   Version: Coq V8.0pl3
   Date: June 2008
                                                                 
   An intuitionistic sequent calculus used as a specification
   logic in the style of FOLDN, defined for use in two-level Hybrid,
   version without natural number arguments
  ***************************************************************)

Require Export List.
Require Export Wf_nat.
Require Export Omega.
Require Export Hybrid.

Section spec.

Set Implicit Arguments.

(****************************************************************
   The intuitionistic object logic.
  ****************************************************************)
Variable atm:Set.
Variable con:Set.

Inductive prp : Set :=
  | atom : atm -> prp
  | T : prp
  | Conj : prp -> prp -> prp
  | Imp : atm -> prp -> prp
  | All : (expr con -> prp) -> prp
  | Some : (expr con -> prp) -> prp.

Variable prog : atm -> prp -> Prop.

Inductive seq : list atm -> prp -> Prop :=
  | sbc :
      forall (A : atm) (L : list atm) (b : prp),
      prog A b -> seq L b -> seq L (atom A)
  | sinit :
      forall (A A' : atm) (L : list atm),
      In A (A' :: L) -> seq (A' :: L) (atom A)
  | strue : forall (L : list atm), seq L T
  | sand :
      forall (B C : prp) (L : list atm),
      seq L B -> seq L C -> seq L (Conj B C)
  | simp :
      forall (A : atm) (B : prp) (L : list atm),
      seq (A :: L) B -> seq L (Imp A B)
  | sall :
      forall (B : expr con -> prp) (L : list atm),
      (forall x : expr con, proper x -> seq L (B x)) -> seq L (All B)
  | ssome :
      forall (B : expr con -> prp) (L : list atm) (x : expr con),
      seq L (B x) -> seq L (Some B).

(* context weakening *)
Theorem seq_thin_exch :
 forall (b : prp) (l l' : list atm),
 (forall a : atm, In a l -> In a l') -> seq l b -> seq l' b.
Proof.
intros b l l' H H0.
generalize H0 l' H; clear H0 H l'.
simple induction 1.
intros A L b0 H H1 H2 l' H3.
apply sbc with b0; auto.
intros A A' L H l' H1.
generalize (H1 A H).
clear H1; generalize l'; clear l'.
induction l'.
intro H1; elim H1; auto.
apply sinit; auto.
intros; apply strue.
intros B C L H H1 H2 H3 l' H4.
apply sand; auto.
intros A B L H H1 l' H2.
apply simp; auto.
apply H1; auto.
intros a H3.
specialize in_inv with (1 := H3); intro H4.
elim H4; clear H4; intro H4.
rewrite H4.
apply in_eq; auto.
apply in_cons; apply H2; auto.
intros B L H H1 l' H2.
apply sall; auto.
intros B L x H H1 l' H2.
apply ssome with x; auto.
Qed.

Theorem specialization :
  forall (l:list atm) (b:expr con -> prp),
  seq l (All b) -> forall (x:expr con), proper x -> seq l (b x).
Proof.
intros l b h0 x h1.
inversion h0; subst.
apply H1; auto.
Qed.

(* atomic cut *)
Theorem seq_cut_aux :
 forall (l' : list atm) (b : prp), seq l' b ->
 forall (l : list atm) (a : atm), In a l' ->
 (forall a' : atm, In a' l' -> a' = a \/ In a' l) ->
 seq l (atom a) -> seq l b.
Proof.
intros l' b; simple induction 1.
intros A L b0 H0 H1 H2 l a H3 H4 H5.
apply sbc with b0; auto.
apply H2 with a; auto.
intros A A' L H0 l a H1 H2 H3.
generalize (H2 A H0); intro H4.
elim H4; clear H4; intro H4.
rewrite H4; auto.
generalize l H3 H4.
induction l0.
intros H5 H6.
inversion H6.
intros H5 H6.
apply sinit; auto.
intros; apply strue.
intros B C L H0 H1 H2 H3 l a H4 H5 H6.
apply sand; auto.
apply H1 with a; auto.
apply H3 with a; auto.
intros A B L H0 H1 l a H2 H3 H4.
apply simp; auto.
apply H1 with a; auto.
apply in_cons; auto.
intros a' H5.
specialize in_inv with (1:=H5); intro H6.
elim H6; clear H6; intro H6.
right; rewrite H6; apply in_eq; auto.
specialize H3 with (1 := H6); intro H7.
elim H7; clear H7; intro H7; auto.
right; apply in_cons; auto.
apply seq_thin_exch with l; auto.
intros a0 H5.
apply in_cons; auto.
intros B L H0 H1 l a H2 H3 H4.
apply sall; auto.
intros x H5.
apply H1 with a; auto.
intros B L x H0 H1 l a H2 H3 H4.
apply ssome with x; auto.
apply H1 with a; auto.
Qed.

Theorem seq_cut :
 forall (a : atm) (b : prp) (l : list atm),
 seq (a :: l) b -> seq l (atom a) -> seq l b.
intros a b l H H0.
apply seq_cut_aux with (a :: l) a; auto.
apply in_eq; auto.
intros a' H1.
specialize in_inv with (1 := H1); auto.
intro H2; elim H2; clear H2; intro H2; auto.
Qed.

Unset Implicit Arguments.

End spec.
