
(****************************************************************
   File: sr_meval.v                                                 
   Authors: Amy Felty
   Version: Coq V8.0pl3
   Date: July 2007                                                                                                                                  
   Subject Reduction for Call-by-Value Mini-ML using Hybrid
   (Version with eval judgments defined as inductive type, and
    no natural number argument in sequents of the specification logic.
    This version uses isterm1 (1-level adequacy) instead of isterm
    (2-level adequacy).
    There is no proof of (typeof M T) implies (proper M).  Such a
    proof requires induction over proofs of isterm judgments, which
    have been left out, and natural number arguments to sequents.
    See sr_nat_meval.)

  ***************************************************************)

Require Import spec.
Require Import miniml.

Section sr.

(****************************************************************
   The atm type and instantiation of prp.
  ****************************************************************)
Inductive atm : Set :=
   typeof : uexp -> tp -> atm.

Definition prp_ := prp atm Lcon.
Definition atom_ : atm -> prp_ := atom Lcon.
Definition T_ : prp_ := T atm Lcon.

(****************************************************************
   Definition of prog
  ****************************************************************)
Inductive prog : atm -> prp_ -> Prop :=
  | tabs :
      forall (T1 T2: tp) (E : uexp -> uexp),
      abstr E ->
      prog (typeof (Fun (fun x => E x)) (arrow T1 T2))
        (All (fun x : uexp => Imp (typeof x T1) (atom_ (typeof (E x) T2))))
  | tap :
      forall E1 E2: uexp, forall T T' : tp,
      prog (typeof (App E1 E2) T)
        (Conj (atom_ (typeof E1 (arrow T' T))) (atom_ (typeof E2 T')))
  | tfix :
      forall (T: tp) (E : uexp -> uexp),
      abstr E ->
      prog (typeof (Fix (fun x => E x)) T)
        (All (fun x : uexp => Imp (typeof x T) (atom_ (typeof (E x) T)))).

(****************************************************************
   Instantiation of seq
  ****************************************************************)

Definition seq_ : list atm -> prp_ -> Prop := seq prog.
Definition sinit_ := sinit prog.
Definition strue_ := strue prog.

Definition seq0 (B : prp_) : Prop := seq_ nil B.

(****************************************************************
   Evaluation
  ****************************************************************)

Inductive eval: uexp -> uexp -> Prop :=
  eval_App: forall E1':uexp -> uexp, forall E1 E2 V V2:uexp,
               abstr E1' ->
               eval E1 (Fun (fun x => E1' x)) ->
               eval E2 V2 -> eval (E1' V2) V ->
               eval (App E1 E2) V
| eval_Fun: forall E:uexp -> uexp, abstr E -> 
               isterm1 nil (Fun (fun x => E x)) ->
               eval (Fun (fun x => E x)) (Fun (fun x => E x))
| eval_Fix: forall E:uexp -> uexp, forall V:uexp,
               abstr E ->
               isterm1 nil (Fix (fun x => E x)) ->
               eval (E (Fix (fun x => E x))) V ->
               eval (Fix (fun x => E x)) V.

(* In Isabelle, this is (eval.mkH_cases (Fun...)) *)
Lemma Fun_elim: forall P:Prop, forall (E:uexp->uexp) (V:uexp),
  (eval (Fun (fun x => E x)) V) ->
  (forall F:uexp->uexp, abstr F -> isterm1 nil (Fun (fun x => F x)) ->
   lambda (fun x => E x)=lambda (fun x => F x) ->
   V=(Fun (fun x => E x)) -> P) -> P.
Proof.
intros P E V H H0.
inversion H; subst.
apply H0 with E0; auto.
fold uexp; unfold lambda.
rewrite H1; auto.
unfold Fun,lambda.
rewrite H1; auto.
Qed.

Lemma eval_unique:
  forall E F:uexp, (eval E F) -> forall G:uexp, (eval E G) -> F=G.
Proof.
intros E F; induction 1.
intros G H2'.
inversion H2'; subst.
generalize (IHeval1 (Fun (fun x : uexp => E1'0 x)) H6); intro H9'.
specialize abstr_eta with (1:=H); intro H10'.
specialize abstr_eta with (1:=H5); intro H11'.
generalize (Fun_inj (fun x => E1' x) (fun x => E1'0 x) H10' H11' H9');
  intro H12'.
unfold ext_eq in H12'.
generalize (IHeval2 V1 H7); intro H7'; subst.
rewrite <- H12' in H9.
apply IHeval3; auto.

intros G h0.
inversion h0; auto.
unfold Fun,lambda.
rewrite H1; auto.

intros G h0.
inversion h0; auto.
specialize abstr_eta with (1:=H); intro H6'.
specialize abstr_eta with (1:=H3); intro H7'.
generalize (abstr_lbind_simp H7' H6' H2); intro H8.
unfold ext_eq in H8.
apply IHeval; auto.
rewrite <- H8; auto.
unfold Fix,lambda.
unfold Fix,lambda in H5.
rewrite <- H2; auto.
Qed.

(****************************************************************
   Adequacy
  ****************************************************************)

Lemma eval_proper_l : forall E V:uexp,
  (eval E V) -> (proper E).
Proof.
intros E V; induction 1.
apply proper_App; auto.
apply proper_Fun; auto.
apply proper_Fix; auto.
Qed.

Lemma eval_proper_r : forall E V:uexp,
  (eval E V) -> (proper V).
Proof.
intros E V; induction 1; auto.
apply proper_Fun; auto.
Qed.

Lemma eval_proper : forall E V:uexp,
  (eval E V) -> (proper E) /\ (proper V).
Proof.
intros E V h1; split.
apply eval_proper_l with V; auto.
apply eval_proper_r with E; auto.
Qed.

Lemma eval_isterm : forall (M N:uexp),
  eval M N -> (isterm1 nil M) /\ (isterm1 nil N).
Proof.
intros M N; induction 1.
(* App case *)
elim IHeval1; intros h1 h2.
elim IHeval2; intros h3 h4.
elim IHeval3; intros h5 h6.
split; auto.
apply term1_app; auto.
(* Fun case *)
split; auto.
(* Fix case *)
elim IHeval; intros h1 h2.
split; auto.
Qed.

(****************************************************************
   Subject Reduction (Type Soundness)
  ****************************************************************)

Theorem meta_subject_reduction :
 forall e v : uexp, (eval e v) ->
 forall t : tp, seq0 (atom_ (typeof e t)) -> seq0 (atom_ (typeof v t)).
Proof.
intros e v; induction 1; auto.
(* App case *)
intros t H3.
unfold seq0 in H3.
inversion H3; clear H3; subst.
inversion H5; clear H5; subst.
inversion H7; clear H7; subst.
assert (seq0 (atom_ (typeof E1 (arrow T' t)))); auto.
assert (seq0 (atom_ (typeof E2 T'))); auto.
clear H6 H8.
specialize IHeval2 with (1:=H4); intro H4'.
specialize IHeval1 with (1 := H3); intro H3'.
unfold seq0 in H3'.
inversion H3'; clear H3'; subst.
inversion H6; clear H6; subst.
inversion H8; clear H8; subst.
generalize (H9 V2 (eval_proper_r E2 V2 H1)); intro H10.
generalize (abstr_eta H); intro H'.
generalize (abstr_eta H11); intro H11'.
generalize (abstr_lbind_simp H11' H' H5); intro H12'.
unfold ext_eq in H12'.
rewrite H12' in H10; clear H12' H11' H5 H11 H9 E.
apply IHeval3; auto.
unfold seq0, seq_.
apply seq_cut with (typeof V2 T'); auto.
inversion H10; clear H10; subst; auto.
(* Fix case *)
intros t H1'.
inversion H1'; subst.
inversion H3; clear H3; subst.
inversion H5; clear H5; subst.
generalize (H6 (Fix (fun x : uexp => E x))); clear H6; intro H5.
generalize (abstr_eta H); intro H5'.
generalize (abstr_eta H7); intro H8'.
generalize (abstr_lbind_simp H8' H5' H2); intro H1''.
unfold ext_eq in H1''.
rewrite -> H1'' in H5.
clear H1'' H8' H2 H7 E0.
generalize (abstr_proper_Fix (fun x => E x) H5'); intro H0'.
generalize (H5 H0'); clear H5; intro H9.
apply IHeval; auto.
unfold seq0, seq_.
apply seq_cut with (typeof (Fix (fun x : uexp => E x)) t); auto.
inversion H9; clear H9; subst; auto.
Qed.

End sr.
