
(****************************************************************
   File: sr_nat.v                                                 
   Authors: Amy Felty
   Version: Coq V8.0pl3
   Date: May 2007
                                                                 
   Subject Reduction for Call-by-Value Mini-ML using Hybrid
   (Version with eval judgments defined using the specification logic.)

  ***************************************************************)

Require Import spec_nat.
Require Import miniml.

Section sr.

(****************************************************************
   The atm type and instantiation of prp.
  ****************************************************************)
Inductive atm : Set :=
   typeof : uexp -> tp -> atm
 | eval : uexp -> uexp -> atm
 | isterm : uexp -> atm.

Definition prp_ := prp atm Lcon.
Definition atom_ : atm -> prp_ := atom Lcon.
Definition T_ : prp_ := T atm Lcon.

(****************************************************************
   Definition of prog
  ****************************************************************)
Inductive prog : atm -> prp_ -> Prop :=
  | tabs :
      forall (T1 T2: tp) (E : uexp -> uexp),
      abstr E ->
      prog (typeof (Fun (fun x => E x)) (arrow T1 T2))
        (All (fun x : uexp => Imp (typeof x T1) (atom_ (typeof (E x) T2))))
  | tap :
      forall E1 E2: uexp, forall T T' : tp,
      prog (typeof (App E1 E2) T)
        (Conj (atom_ (typeof E1 (arrow T' T))) (atom_ (typeof E2 T')))
  | tfix :
      forall (T: tp) (E : uexp -> uexp),
      abstr E ->
      prog (typeof (Fix (fun x => E x)) T)
        (All (fun x : uexp => Imp (typeof x T) (atom_ (typeof (E x) T))))
  | eap : forall E1':uexp -> uexp, forall E1 E2 V V2:uexp,
      abstr E1' ->
      prog (eval (App E1 E2) V)
        (Conj (atom_ (eval E1 (Fun (fun x => E1' x))))
         (Conj (atom_ (eval E2 V2))
          (atom_ (eval (E1' V2) V))))
  | eabs : forall E:uexp -> uexp, (abstr E) -> 
      prog (eval (Fun (fun x => E x)) (Fun (fun x => E x)))
        (atom_ (isterm (Fun (fun x => E x))))
  | efix : forall E:uexp -> uexp, forall V:uexp,
      abstr E ->
      prog (eval (Fix (fun x => E x)) V)
        (Conj (atom_ (eval (E (Fix (fun x => E x))) V))
         (atom_ (isterm (Fix (fun x => E x)))))
  | exp_app : forall M N:uexp,
      prog (isterm (App M N))
        (Conj (atom_ (isterm M)) (atom_ (isterm N)))
  | exp_lam : forall (E:uexp -> uexp),
      abstr E ->
      prog (isterm (Fun (fun x => (E x))))
        (All (fun x:uexp => (Imp (isterm x) (atom_ (isterm (E x))))))
  | exp_fix : forall (E:uexp -> uexp),
      abstr E ->
      prog (isterm (Fix (fun x => (E x))))
        (All (fun x:uexp => (Imp (isterm x) (atom_ (isterm (E x)))))).
          
(****************************************************************
   Instantiation of seq
  ****************************************************************)

Definition seq_ : nat -> list atm -> prp_ -> Prop := seq prog.
Definition sinit_ := sinit prog.
Definition strue_ := strue prog.

Definition seq'_ := seq' prog.

Definition seq0 (B : prp_) : Prop := exists i : nat, seq_ i nil B.

(****************************************************************
   Example
  ****************************************************************)

Lemma test2 : exists t:tp, seq0 (atom_ (typeof example2 t)).
Proof.
unfold example2.
exists (arrow gnd (arrow gnd gnd)).
unfold seq0.
exists (0+1+1+1+1+1+1).
unfold seq_,atom_.
apply sbc with (All (fun x => Imp (typeof x gnd)
                      (atom_ (typeof (Fun (fun _ => x)) (arrow gnd gnd))))).
generalize (tabs gnd (arrow gnd gnd)
             (fun x : uexp => Fun (fun _ : uexp => x))); intro H.
apply H; auto.
apply test2_abstr.
apply sall.
intros x H.
apply simp.
unfold atom_.
apply sbc with (All (fun y => Imp (typeof y gnd) (atom_ (typeof x gnd)))).
generalize (tabs gnd gnd (fun y => x)); intro H0.
apply H0; auto.
unfold uexp; apply proper_abstr; auto.
apply sall.
intros x0 H0.
apply simp.
apply sinit_; auto.
simpl; tauto.
Qed.

(****************************************************************
   Properties of prog
  ****************************************************************)

Lemma typeof_abs_inv :
forall (i:nat) (l:list atm) (T T':tp) (E:uexp->uexp),
(forall x : uexp,
       proper x ->
       seq_ i l (Imp (typeof x T) (atom_ (typeof (E x) T')))) ->
exists j:nat, (i=j+1 /\ 
 forall x : uexp,
       proper x ->
       seq_ j ((typeof x T)::l) (atom_ (typeof (E x) T'))).
Proof.
induction i; auto.
intros l T T' E H.
generalize (H (Var 0) (proper_Var 0)); intro H1.
inversion H1; clear H1; subst.
replace (i+1) with (S i) in H0; try omega.
inversion H0.
intros l T T' E H.
clear IHi.
exists i; split; try omega.
intros x H0.
generalize (H x H0); intro H1.
inversion H1; clear H1; subst.
assert (i0 = i); try omega.
subst; auto.
Qed.

(****************************************************************
   Adequacy
  ****************************************************************)

Lemma eval_proper_l : forall E V:uexp,
  seq0 (atom_ (eval E V)) -> (proper E).
Proof.
intros E V [n h1].
generalize E V h1; clear h1 E V.
generalize
 (lt_wf_ind n
    (fun n:nat =>
       forall E V : uexp, seq_ n nil (atom_ (eval E V)) -> proper E)).
intro h'.
apply h'; clear h'; auto.
clear n.
intros n h1 E V h2.
inversion h2; clear h2; subst.
inversion H0; clear H0; subst.
inversion H3; clear H3; subst.
inversion H6; clear H6; subst.
assert (i<i+1+1+1); try omega.
assert (i+1<i+1+1+1); try omega.
generalize (h1 i H E2 V2 H3); intro h2.
generalize (h1 (i+1) H0 E1 (Fun (fun x : uexp => E1' x)) H5); intro h3.
apply proper_App; auto.
apply proper_Fun; auto.
apply proper_Fix; auto.
Qed.

Lemma eval_proper_r : forall E V:uexp,
  seq0 (atom_ (eval E V)) -> (proper V).
intros E V [n h1].
generalize E V h1; clear h1 E V.
generalize
 (lt_wf_ind n
    (fun n:nat =>
       forall E V : uexp, seq_ n nil (atom_ (eval E V)) -> proper V)).
intro h'.
apply h'; clear h'; auto.
clear n.
intros n h1 E V h2.
inversion h2; clear h2; subst.
inversion H0; clear H0; subst.
inversion H3; clear H3; subst.
inversion H6; clear H6; subst.
apply h1 with i (E1' V2); auto; omega.
apply proper_Fun; auto.
inversion H3; clear H3; subst.
apply h1 with i0 (E0 (Fix (fun x : uexp => E0 x))); auto; omega.
Qed.

Lemma eval_proper : forall E V:uexp,
  seq0 (atom_ (eval E V)) -> (proper E) /\ (proper V).
Proof.
intros E V h1; split.
apply eval_proper_l with V; auto.
apply eval_proper_r with E; auto.
Qed.

Lemma hastype_proper : forall (E:uexp) (T:tp),
  seq0 (atom_ (typeof E T)) -> (proper E).
Proof.
intros E T [n h1].
generalize E T h1; clear h1 E T.
generalize
 (lt_wf_ind n
    (fun n:nat =>
     forall (E : uexp) (T : tp), seq_ n nil (atom_ (typeof E T)) -> proper E)).
intro h'.
apply h'; clear h'; auto.
clear n.
intros n h1 E T h2.
inversion h2; clear h2; subst.
inversion H0; clear H0; subst.
apply proper_Fun; auto.
inversion H3; clear H3; subst.
apply proper_App; auto.
apply h1 with i0 (arrow T' T); auto; try omega.
apply h1 with i0 T'; auto; try omega.
apply proper_Fix; auto.
Qed.

Lemma hastype_isterm_aux :
  forall (i:nat) (M:uexp) (T:tp) (l l':list atm),
  (forall (M:uexp) (T:tp), (In (typeof M T) l) -> (In (isterm M) l')) ->
  seq_ i l (atom_ (typeof M T)) ->
  seq_ i l' (atom_ (isterm M)).
Proof.
intro i.
generalize
 (lt_wf_ind i
    (fun i:nat =>
     forall (M : uexp) (T : tp) (l l' : list atm),
     (forall (M0 : uexp) (T0 : tp), In (typeof M0 T0) l -> In (isterm M0) l') ->
     seq_ i l (atom_ (typeof M T)) -> seq_ i l' (atom_ (isterm M)))).
intro H'.
apply H'; clear H' i; auto.
intros i h M T l l' h1 h2.
inversion h2; subst; clear h2.
inversion H0; subst; clear H0.
(* Fun case *)
inversion H3; subst; clear H3.
generalize (typeof_abs_inv i l T1 T2 E H2); clear H2; intros [j [h3 h4]];
 subst.
unfold seq_,atom_; apply sbc with
 (All (fun x:uexp => (Imp (isterm x) (atom_ (isterm (E x)))))).
apply exp_lam; auto.
apply sall; auto.
intros x h5.
apply simp; auto.
apply h with T2 (typeof x T1::l); auto.
omega.
intros M T; simpl.
intros [h6 | h6].
injection h6; intros h7 h8; subst.
left; auto.
right; apply h1 with T; auto.
(* App case *)
inversion H3; subst; clear H3.
unfold seq_,atom_; apply sbc with (Conj (atom_ (isterm E1)) (atom_ (isterm E2))).
apply exp_app.
apply sand; auto.
apply h with (arrow T' T) l; auto.
omega.
apply h with T' l; auto.
omega.
(* Fix case *)
inversion H3; subst; clear H3.
generalize (typeof_abs_inv i l T T E H2); clear H2; intros [j [h3 h4]];
 subst.
unfold seq_,atom_; apply sbc with
 (All (fun x:uexp => (Imp (isterm x) (atom_ (isterm (E x)))))).
apply exp_fix; auto.
apply sall; auto.
intros x h5.
apply simp; auto.
apply h with T (typeof x T::l); auto.
omega.
intros M T'; simpl.
intros [h6 | h6].
injection h6; intros h7 h8; subst.
left; auto.
right; apply h1 with T'; auto.
(* context case *)
specialize h1 with (1:=H2); intro h3.
generalize h3; clear h3 h1.
induction l'.
simpl; tauto.
unfold seq_,atom_; apply sinit.
Qed.

Lemma hastype_isterm : forall (M:uexp) (T:tp),
  seq0 (atom_ (typeof M T)) ->
  seq0 (atom_ (isterm M)).
Proof.
intros M T [i h].
exists i.
generalize (hastype_isterm_aux i M T nil nil); intro h2.
apply h2; auto.
Qed.

Lemma eval_isterm_aux : forall (i:nat) (M N:uexp),
  seq_ i nil (atom_ (eval M N)) ->
  (seq_ i nil (atom_ (isterm M)) /\ seq_ i nil (atom_ (isterm N))).
Proof.
intro i.
generalize
 (lt_wf_ind i
    (fun i:nat =>
     forall M N : uexp,
     seq_ i nil (atom_ (eval M N)) ->
     seq_ i nil (atom_ (isterm M)) /\ seq_ i nil (atom_ (isterm N)))).
intro H'; apply H'; clear H' i; auto.
intros i h M N h3.
inversion h3; subst; clear h3.
inversion H0; subst; clear H0.
inversion H3; subst; clear H3.
inversion H6; subst; clear H6.
(* App case *)
assert (ha1: (i0<i0+1+1+1)); try omega.
assert (ha2: (i0+1<i0+1+1+1)); try omega.
specialize h with (1:=ha1) (2:=H3); intros [h1 h2].
specialize h with (1:=ha1) (2:=H7); intros [h3 h4].
specialize h with (1:=ha2) (2:=H5); intros [h5 h6].
split.
unfold seq_,atom_; apply sbc with
 (Conj (atom_ (isterm E1)) (atom_ (isterm E2))).
apply exp_app.
apply sand; auto.
apply seq_mono with i0; auto; try omega.
unfold seq_,atom_; apply seq_mono with i0; auto; try omega.
unfold seq_,atom_; split; apply seq_mono with i0; auto; try omega.
inversion H3; subst; clear H3.
split.
unfold seq_,atom_; apply seq_mono with i; auto; try omega.
assert (ha:(i<i+1+1)); try omega.
specialize h with (1:=ha) (2:=H5); intros [h1 h2].
unfold seq_,atom_; apply seq_mono with i; auto; try omega.
Qed.

Lemma eval_isterm : forall (M N:uexp),
  seq0 (atom_ (eval M N)) ->
  (seq0 (atom_ (isterm M)) /\ seq0 (atom_ (isterm N))).
Proof.
intros M N [i h].
specialize eval_isterm_aux with (1:=h); intros [h1 h2].
split; exists i; auto.
Qed.

(****************************************************************
   Subject Reduction (Type Soundness)
  ****************************************************************)

Theorem OL_subject_reduction :
 forall e v : uexp, seq0 (atom_ (eval e v)) ->
 forall t : tp, seq0 (atom_ (typeof e t)) -> seq0 (atom_ (typeof v t)).
intros e v H t H0.
unfold seq0 in H; elim H; clear H; intros n H.
generalize
 (lt_wf_ind n
    (fun j:nat =>
       forall e v:uexp, seq_ j nil (atom_ (eval e v)) ->
       forall t:tp, seq0 (atom_ (typeof e t)) ->
       seq0 (atom_ (typeof v t)))).
intro H'.
apply H' with e; clear H'; auto.
clear H n H0 e v t.
intros n H e v H0 t H1.
inversion H0; clear H0; subst.
inversion H3; clear H3; subst.
(* App case *)
inversion H6; clear H6; subst.
inversion H8; clear H8; subst.
unfold seq0 in H1; elim H1; clear H1; intros j H1.
inversion H1; clear H1; subst.
inversion H2; clear H2; subst.
inversion H8; clear H8; subst.
assert (seq0 (atom_ (typeof E1 (arrow T' t)))).
unfold seq0,seq_; exists i1; auto.
assert (seq0 (atom_ (typeof E2 T'))).
unfold seq0,seq_; exists i1; auto.
clear H4 H10.
assert (i < i+1+1+1); try omega.
assert (i+1 < i+1+1+1); try omega.
generalize (H i H2 E2 V2 H6 T' H1); intro H4.
generalize (H (i+1) H3 E1 (Fun (fun x => E1' x)) H7 (arrow T' t) H0);
 intro H8.
unfold seq0 in H8.
elim H8; clear H8; intros j H8.
inversion H8; clear H8; subst.
inversion H11; clear H11; subst.
inversion H14; clear H14; subst.
generalize (typeof_abs_inv i2 nil T' t E H13); intro H10.
elim H10; clear H10; intros j [Heq H10].
subst.
clear H13.
assert (proper V2).
apply eval_proper_r with E2.
unfold seq0; exists i; auto.
generalize (H10 V2 H11); clear H10; intro H10.
generalize (abstr_eta H15); intro H15'.
generalize (abstr_eta H5); intro H5'.
generalize (abstr_lbind_simp H15' H5' H8); intro H12.
unfold ext_eq in H12.
rewrite H12 in H10; clear H12 H15' H8 H15 E.
apply H with i (E1' V2); auto; try omega.
unfold seq0, seq_.
apply seq_cut with (typeof V2 T'); auto.
unfold seq'.
exists j; auto.
(* Fun case *)
exact H1.
(* Fix case *)
elim H1; intros j H1'.
inversion H1'; clear H1'; subst.
inversion H2; clear H2; subst.
inversion H7; clear H7; subst.
generalize (typeof_abs_inv i1 nil t t E0 H9); clear H9; intro H9.
elim H9; clear H9; intros j [Heq H9]; subst.
generalize (H9 (Fix (fun x : uexp => E x))); clear H9; intro H9.
generalize (abstr_eta H5); intro H5'.
generalize (abstr_eta H8); intro H8'.
generalize (abstr_lbind_simp H8' H5' H0); intro H1'.
unfold ext_eq in H1'.
rewrite -> H1' in H9.
clear H1' H8' H0 H8 E0.
generalize (abstr_proper_Fix (fun x => E x) H5'); intro H0.
generalize (H9 H0); clear H9; intro H9.
inversion H6; subst; clear H6.
apply H with i0 (E (Fix (fun x : uexp => E x))); auto; try omega.
unfold seq0, seq_.
apply seq_cut with (typeof (Fix (fun x : uexp => E x)) t); auto.
unfold seq'.
exists j; auto.
Qed.

End sr.
