
(****************************************************************
   File: sr_nat_meval.v
   Authors: Amy Felty
   Version: Coq V8.0pl3
   Date: July 2007

   Subject Reduction for Call-by-Value Mini-ML using Hybrid
   (Version with eval judgments defined as inductive type.
    This version uses isterm (2-level adequacy).)

  ***************************************************************)

Require Import spec_nat.
Require Import miniml.

Section sr.

(****************************************************************
   The atm type and instantiation of prp.
  ****************************************************************)
Inductive atm : Set :=
   typeof : uexp -> tp -> atm
 | isterm : uexp -> atm.

Definition prp_ := prp atm Lcon.
Definition atom_ : atm -> prp_ := atom Lcon.
Definition T_ : prp_ := T atm Lcon.

(****************************************************************
   Definition of prog
  ****************************************************************)
Inductive prog : atm -> prp_ -> Prop :=
  | tabs :
      forall (T1 T2: tp) (E : uexp -> uexp),
      abstr E ->
      prog (typeof (Fun (fun x => E x)) (arrow T1 T2))
        (All (fun x : uexp => Imp (typeof x T1) (atom_ (typeof (E x) T2))))
  | tap :
      forall E1 E2: uexp, forall T T' : tp,
      prog (typeof (App E1 E2) T)
        (Conj (atom_ (typeof E1 (arrow T' T))) (atom_ (typeof E2 T')))
  | tfix :
      forall (T: tp) (E : uexp -> uexp),
      abstr E ->
      prog (typeof (Fix (fun x => E x)) T)
        (All (fun x : uexp => Imp (typeof x T) (atom_ (typeof (E x) T))))
  | exp_app : forall M N:uexp,
      prog (isterm (App M N))
        (Conj (atom_ (isterm M)) (atom_ (isterm N)))
  | exp_lam : forall (E:uexp -> uexp),
      abstr E ->
      prog (isterm (Fun (fun x => (E x))))
        (All (fun x:uexp => (Imp (isterm x) (atom_ (isterm (E x))))))
  | exp_fix : forall (E:uexp -> uexp),
      abstr E ->
      prog (isterm (Fix (fun x => (E x))))
        (All (fun x:uexp => (Imp (isterm x) (atom_ (isterm (E x)))))).
          
(****************************************************************
   Instantiation of seq
  ****************************************************************)

Definition seq_ : nat -> list atm -> prp_ -> Prop := seq prog.
Definition sinit_ := sinit prog.
Definition strue_ := strue prog.

Definition seq'_ := seq' prog.

Definition seq0 (B : prp_) : Prop := exists i : nat, seq_ i nil B.

(****************************************************************
   Properties of prog
  ****************************************************************)

Lemma typeof_abs_inv :
forall (i:nat) (l:list atm) (T T':tp) (E:uexp->uexp),
(forall x : uexp,
       proper x ->
       seq_ i l (Imp (typeof x T) (atom_ (typeof (E x) T')))) ->
exists j:nat, (i=j+1 /\ 
 forall x : uexp,
       proper x ->
       seq_ j ((typeof x T)::l) (atom_ (typeof (E x) T'))).
Proof.
induction i; auto.
intros l T T' E H.
generalize (H (Var 0) (proper_Var 0)); intro H1.
inversion H1; clear H1; subst.
replace (i+1) with (S i) in H0; try omega.
inversion H0.
intros l T T' E H.
clear IHi.
exists i; split; try omega.
intros x H0.
generalize (H x H0); intro H1.
inversion H1; clear H1; subst.
assert (i0 = i); try omega.
subst; auto.
Qed.

(****************************************************************
   Evaluation
  ****************************************************************)

Inductive eval: uexp -> uexp -> Prop :=
  eval_App: forall E1':uexp -> uexp, forall E1 E2 V V2:uexp,
               abstr E1' ->
               eval E1 (Fun (fun x => E1' x)) ->
               eval E2 V2 -> eval (E1' V2) V ->
               eval (App E1 E2) V
| eval_Fun: forall E:uexp -> uexp, abstr E -> 
               seq0 (atom_ (isterm (Fun (fun x => E x)))) ->
               eval (Fun (fun x => E x)) (Fun (fun x => E x))
| eval_Fix: forall E:uexp -> uexp, forall V:uexp,
               abstr E ->
               seq0 (atom_ (isterm (Fix (fun x => E x)))) ->
               eval (E (Fix (fun x => E x))) V ->
               eval (Fix (fun x => E x)) V.

(* In Isabelle, this is (eval.mkH_cases (Fun...)) *)
Lemma Fun_elim: forall P:Prop, forall (E:uexp->uexp) (V:uexp),
  (eval (Fun (fun x => E x)) V) ->
  (forall F:uexp->uexp, abstr F ->
   seq0 (atom_ (isterm (Fun (fun x => F x)))) ->
   lambda (fun x => E x)=lambda (fun x => F x) ->
   V=(Fun (fun x => E x)) -> P) -> P.
Proof.
intros P E V H H0.
inversion H; subst.
apply H0 with E0; auto.
fold uexp; unfold lambda.
rewrite H1; auto.
unfold Fun,lambda.
rewrite H1; auto.
Qed.

Lemma eval_unique:
  forall E F:uexp, (eval E F) -> forall G:uexp, (eval E G) -> F=G.
Proof.
intros E F; induction 1.
intros G H2'.
inversion H2'; subst.
generalize (IHeval1 (Fun (fun x : uexp => E1'0 x)) H6); intro H9'.
specialize abstr_eta with (1:=H); intro H10'.
specialize abstr_eta with (1:=H5); intro H11'.
generalize (Fun_inj (fun x => E1' x) (fun x => E1'0 x) H10' H11' H9');
  intro H12'.
unfold ext_eq in H12'.
generalize (IHeval2 V1 H7); intro H7'; subst.
rewrite <- H12' in H9.
apply IHeval3; auto.

intros G h0.
inversion h0; auto.
unfold Fun,lambda.
rewrite H1; auto.

intros G H1'.
inversion H1'; auto.
specialize abstr_eta with (1:=H); intro H6'.
specialize abstr_eta with (1:=H3); intro H7'.
generalize (abstr_lbind_simp H7' H6' H2); intro H8.
unfold ext_eq in H8.
apply IHeval; auto.
rewrite <- H8; auto.
unfold Fix,lambda.
unfold Fix,lambda in H0.
rewrite <- H2; auto.
Qed.

(****************************************************************
   Adequacy
  ****************************************************************)

Lemma eval_proper_l : forall E V:uexp,
  (eval E V) -> (proper E).
Proof.
intros E V; induction 1.
apply proper_App; auto.
apply proper_Fun; auto.
apply proper_Fix; auto.
Qed.

Lemma eval_proper_r : forall E V:uexp,
  (eval E V) -> (proper V).
Proof.
intros E V; induction 1; auto.
apply proper_Fun; auto.
Qed.

Lemma eval_proper : forall E V:uexp,
  (eval E V) -> (proper E) /\ (proper V).
Proof.
intros E V h1; split.
apply eval_proper_l with V; auto.
apply eval_proper_r with E; auto.
Qed.

Lemma hastype_proper : forall (E:uexp) (T:tp),
  seq0 (atom_ (typeof E T)) -> (proper E).
Proof.
intros E T [n h1].
generalize E T h1; clear h1 E T.
generalize
 (lt_wf_ind n
    (fun n:nat =>
     forall (E : uexp) (T : tp), seq_ n nil (atom_ (typeof E T)) -> proper E)).
intro h'.
apply h'; clear h'; auto.
clear n.
intros n h1 E T h2.
inversion h2; clear h2; subst.
inversion H0; clear H0; subst.
apply proper_Fun; auto.
inversion H3; clear H3; subst.
apply proper_App; auto.
apply h1 with i0 (arrow T' T); auto; try omega.
apply h1 with i0 T'; auto; try omega.
apply proper_Fix; auto.
Qed.

Lemma hastype_isterm_aux :
  forall (i:nat) (M:uexp) (T:tp) (l l':list atm),
  (forall (M:uexp) (T:tp), (In (typeof M T) l) -> (In (isterm M) l')) ->
  seq_ i l (atom_ (typeof M T)) ->
  seq_ i l' (atom_ (isterm M)).
Proof.
intro i.
generalize
 (lt_wf_ind i
    (fun i:nat =>
     forall (M : uexp) (T : tp) (l l' : list atm),
     (forall (M0 : uexp) (T0 : tp), In (typeof M0 T0) l -> In (isterm M0) l') ->
     seq_ i l (atom_ (typeof M T)) -> seq_ i l' (atom_ (isterm M)))).
intro H'.
apply H'; clear H' i; auto.
intros i h M T l l' h1 h2.
inversion h2; subst; clear h2.
inversion H0; subst; clear H0.
(* Fun case *)
inversion H3; subst; clear H3.
generalize (typeof_abs_inv i l T1 T2 E H2); clear H2; intros [j [h3 h4]];
 subst.
unfold seq_,atom_; apply sbc with
 (All (fun x:uexp => (Imp (isterm x) (atom_ (isterm (E x)))))).
apply exp_lam; auto.
apply sall; auto.
intros x h5.
apply simp; auto.
apply h with T2 (typeof x T1::l); auto.
omega.
intros M T; simpl.
intros [h6 | h6].
injection h6; intros h7 h8; subst.
left; auto.
right; apply h1 with T; auto.
(* App case *)
inversion H3; subst; clear H3.
unfold seq_,atom_; apply sbc with (Conj (atom_ (isterm E1)) (atom_ (isterm E2))).
apply exp_app.
apply sand; auto.
apply h with (arrow T' T) l; auto.
omega.
apply h with T' l; auto.
omega.
(* Fix case *)
inversion H3; subst; clear H3.
generalize (typeof_abs_inv i l T T E H2); clear H2; intros [j [h3 h4]];
 subst.
unfold seq_,atom_; apply sbc with
 (All (fun x:uexp => (Imp (isterm x) (atom_ (isterm (E x)))))).
apply exp_fix; auto.
apply sall; auto.
intros x h5.
apply simp; auto.
apply h with T (typeof x T::l); auto.
omega.
intros M T'; simpl.
intros [h6 | h6].
injection h6; intros h7 h8; subst.
left; auto.
right; apply h1 with T'; auto.
(* context case *)
specialize h1 with (1:=H2); intro h3.
generalize h3; clear h3 h1.
induction l'.
simpl; tauto.
unfold seq_,atom_; apply sinit.
Qed.

Lemma hastype_isterm : forall (M:uexp) (T:tp),
  seq0 (atom_ (typeof M T)) ->
  seq0 (atom_ (isterm M)).
Proof.
intros M T [i h].
exists i.
generalize (hastype_isterm_aux i M T nil nil); intro h2.
apply h2; auto.
Qed.

Lemma eval_isterm : forall (M N:uexp),
  eval M N ->
  (seq0 (atom_ (isterm M)) /\ seq0 (atom_ (isterm N))).
Proof.
intros M N; induction 1.
(* App case *)
elim IHeval1; intros [i1 h1] [i2 h2].
elim IHeval2; intros [i3 h3] [i4 h4].
elim IHeval3; intros [i5 h5] [i6 h6].
split.
exists (i1+i3+1+1).
unfold seq_,atom_; apply sbc with
 (Conj (atom_ (isterm E1)) (atom_ (isterm E2))).
apply exp_app.
apply sand; auto.
unfold atom_; apply seq_mono_cor with i1; auto; try omega.
unfold atom_; apply seq_mono_cor with i3; auto; try omega.
exists i6; auto.
(* Fun case *)
split; auto.
(* Fix case *)
elim IHeval; intros h1 h2.
split; auto.
Qed.

(****************************************************************
   Subject Reduction (Type Soundness)
  ****************************************************************)

Theorem meta_subject_reduction :
 forall e v : uexp, (eval e v) ->
 forall t : tp, seq0 (atom_ (typeof e t)) -> seq0 (atom_ (typeof v t)).
intros e v; induction 1.
(* App case *)
intros t H3.
unfold seq0 in H3; elim H3; clear H3; intros j H3.
inversion H3; clear H3; subst.
inversion H5; clear H5; subst.
inversion H8; clear H8; subst.
assert (seq0 (atom_ (typeof E1 (arrow T' t)))).
unfold seq0,seq_; exists i0; auto.
assert (seq0 (atom_ (typeof E2 T'))).
unfold seq0,seq_; exists i0; auto.
clear H7 H9.
generalize (IHeval2 T' H4); intro H4'.
generalize (IHeval1 (arrow T' t) H3); intro H8.
unfold seq0 in H8; elim H8; clear H8; intros j H8.
inversion H8; clear H8; subst.
inversion H6; clear H6; subst.
inversion H10; clear H10; subst.
generalize (typeof_abs_inv i1 nil T' t E H9); intro H10.
elim H10; clear H10; intros j [Heq H10].
subst.
clear H9.
generalize (H10 V2 (eval_proper_r E2 V2 H1)); clear H10; intro H10.
generalize (abstr_eta H); intro H15'.
generalize (abstr_eta H11); intro H5'.
generalize (abstr_lbind_simp H5' H15' H5); intro H12.
unfold ext_eq in H12.
rewrite H12 in H10; clear H12 H5' H5 H11 E.
apply IHeval3; auto.
unfold seq0, seq_.
apply seq_cut with (typeof V2 T'); auto.
unfold seq'.
exists j; auto.
(* Fun case *)
auto.
(* Fix case *)
intros t h1.
elim h1; intros j H1'.
inversion H1'; clear H1'; subst.
inversion H3; clear H3; subst.
inversion H6; clear H6; subst.
generalize (typeof_abs_inv i0 nil t t E0 H8); clear H8; intro H9.
elim H9; clear H9; intros j [Heq H9]; subst.
generalize (H9 (Fix (fun x : uexp => E x))); clear H9; intro H9.
generalize (abstr_eta H); intro H5'.
generalize (abstr_eta H7); intro H8'.
generalize (abstr_lbind_simp H8' H5' H2); intro H1'.
unfold ext_eq in H1'.
rewrite -> H1' in H9.
clear H1' H8' H2 H7 E0.
generalize (abstr_proper_Fix (fun x => E x) H5'); intro H0'.
generalize (H9 H0'); clear H9; intro H9.
apply IHeval; auto.
unfold seq0, seq_.
apply seq_cut with (typeof (Fix (fun x : uexp => E x)) t); auto.
unfold seq'.
exists j; auto.
Qed.

End sr.
